"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var utils_1 = require("./utils");
var matchesState_1 = require("./matchesState");
var State_1 = require("./State");
var actions_1 = require("./actions");
var STATE_DELIMITER = '.';
var HISTORY_KEY = '$history';
var NULL_EVENT = '';
var StateNode = /** @class */ (function () {
    function StateNode(config) {
        var _this = this;
        this.config = config;
        this.__cache = {
            events: undefined,
            relativeValue: new Map(),
            initialState: undefined
        };
        this.key = config.key || '(machine)';
        this.parent = config.parent;
        this.machine = this.parent ? this.parent.machine : this;
        this.id = this.parent
            ? this.parent.path.concat(this.key).join(STATE_DELIMITER)
            : this.key;
        this.path = this.parent ? this.parent.path.concat(this.key) : [this.key];
        this.initial = config.initial;
        this.parallel = !!config.parallel;
        this.states = (config.states
            ? utils_1.mapValues(config.states, function (stateConfig, key) {
                return new StateNode(__assign({}, stateConfig, { key: key, parent: _this }));
            })
            : {});
        this.on = config.on;
        this.strict = !!config.strict;
        this.onEntry = config.onEntry
            ? [].concat(config.onEntry)
            : undefined;
        this.onExit = config.onExit
            ? [].concat(config.onExit)
            : undefined;
        this.data = config.data;
        this.activities = config.activities;
    }
    StateNode.prototype.getStateNodes = function (state) {
        var _this = this;
        var stateValue = state instanceof State_1.State ? state.value : utils_1.toStateValue(state);
        if (typeof stateValue === 'string') {
            var initialStateValue = this.states[stateValue].initial;
            return initialStateValue
                ? this.getStateNodes((_a = {}, _a[stateValue] = initialStateValue, _a))
                : [this.states[stateValue]];
        }
        var subStateKeys = Object.keys(stateValue);
        var subStateNodes = subStateKeys.map(function (subStateKey) { return _this.states[subStateKey]; });
        return subStateNodes.concat(subStateKeys
            .map(function (subStateKey) {
            return _this.states[subStateKey].getStateNodes(stateValue[subStateKey]);
        })
            .reduce(function (a, b) { return a.concat(b); }));
        var _a;
    };
    StateNode.prototype.handles = function (event) {
        var eventType = utils_1.getEventType(event);
        return this.events.indexOf(eventType) !== -1;
    };
    StateNode.prototype.transition = function (state, event, extendedState) {
        if (this.strict) {
            var eventType = utils_1.getEventType(event);
            if (this.events.indexOf(eventType) === -1) {
                throw new Error("Machine '" + this.id + "' does not accept event '" + eventType + "'");
            }
        }
        var currentState = State_1.State.from(state);
        var stateTransition = this.transitionStateValue(currentState, event, currentState, extendedState);
        var nextState = this.stateTransitionToState(stateTransition, currentState);
        if (!nextState) {
            return State_1.State.inert(currentState);
        }
        // Try transitioning from "transient" states from the NULL_EVENT
        // until a state with non-transient transitions is found
        var maybeNextState = nextState;
        var nextStateFromInternalTransition;
        var actions = [];
        do {
            actions.push.apply(actions, maybeNextState.actions);
            nextStateFromInternalTransition = maybeNextState;
            maybeNextState = this.stateTransitionToState(this.transitionStateValue(nextStateFromInternalTransition, NULL_EVENT, maybeNextState, extendedState), nextStateFromInternalTransition);
        } while (maybeNextState);
        nextStateFromInternalTransition.actions = actions;
        // TODO: handle internally raised events
        return nextStateFromInternalTransition;
    };
    StateNode.prototype.stateTransitionToState = function (stateTransition, prevState) {
        var nextStateValue = stateTransition.stateValue, nextActions = stateTransition.actions, nextActivities = stateTransition.activities;
        if (!nextStateValue) {
            return undefined;
        }
        var prevActivities = prevState instanceof State_1.State ? prevState.activities : undefined;
        var activities = __assign({}, prevActivities, nextActivities);
        return new State_1.State(
        // next state value
        nextStateValue, 
        // history
        State_1.State.from(prevState), 
        // effects
        nextActions
            ? nextActions.onExit
                .concat(nextActions.actions)
                .concat(nextActions.onEntry)
            : [], 
        // activities
        activities, 
        // data
        this.getStateNodes(nextStateValue).reduce(function (data, stateNode) {
            if (stateNode.data !== undefined) {
                data[stateNode.id] = stateNode.data;
            }
            return data;
        }, {}));
    };
    StateNode.prototype.transitionStateValue = function (state, event, fullState, extendedState) {
        var _this = this;
        var history = state.history;
        var stateValue = state.value;
        if (typeof stateValue === 'string') {
            if (!this.states[stateValue]) {
                throw new Error("State '" + stateValue + "' does not exist on machine '" + this.id + "'");
            }
            var subStateNode = this.states[stateValue];
            stateValue = subStateNode.resolvedStateValue;
            if (typeof stateValue === 'string') {
                // Transition from the substate
                return subStateNode.next(event, fullState, history ? history.value : undefined, extendedState);
            }
        }
        // Potential transition tuples from parent state nodes
        var potentialStateTransitions = [];
        var nextStateValue = utils_1.mapValues(stateValue, function (subStateValue, subStateKey) {
            var subHistory = history ? history.value[subStateKey] : undefined;
            var subState = new State_1.State(subStateValue, subHistory ? State_1.State.from(subHistory) : undefined);
            var subStateNode = _this.states[subStateKey];
            var subStateTransition = subStateNode.transitionStateValue(subState, event, fullState, extendedState);
            if (!subStateTransition.stateValue) {
                potentialStateTransitions.push(subStateNode.next(event, fullState, history ? history.value : undefined, extendedState));
            }
            return subStateTransition;
        });
        if (Array.prototype.every.call(Object.keys(nextStateValue), function (key) {
            return nextStateValue[key].stateValue === undefined;
        })) {
            if (this.parallel) {
                if (potentialStateTransitions.length) {
                    return potentialStateTransitions[0];
                }
                return {
                    stateValue: undefined,
                    actions: { onEntry: [], onExit: [], actions: [] },
                    activities: undefined
                };
            }
            var subStateKey = Object.keys(nextStateValue)[0];
            // try with parent
            var _a = this.states[subStateKey].next(event, fullState, history ? history.value : undefined, extendedState), parentNextValue = _a.stateValue, parentNextActions = _a.actions, parentActivities = _a.activities;
            var nextActions = nextStateValue[subStateKey].actions;
            var activities = nextStateValue[subStateKey].activities;
            var allActivities = __assign({}, activities, parentActivities);
            var allActions = parentNextActions
                ? nextActions
                    ? {
                        onEntry: nextActions.onEntry.concat(parentNextActions.onEntry),
                        actions: nextActions.actions.concat(parentNextActions.actions),
                        onExit: nextActions.onExit.concat(parentNextActions.onExit)
                    }
                    : parentNextActions
                : nextActions;
            return {
                stateValue: parentNextValue,
                actions: allActions,
                activities: allActivities
            };
        }
        if (this.parallel) {
            nextStateValue = __assign({}, utils_1.mapValues(this.initialState.value, function (subStateValue) { return ({
                stateValue: subStateValue,
                actions: { onEntry: [], onExit: [], actions: [] },
                activities: undefined
            }); }), nextStateValue);
        }
        var finalActions = {
            onEntry: [],
            actions: [],
            onExit: []
        };
        var finalActivities = {};
        var finalStateValue = utils_1.mapValues(nextStateValue, function (subStateTransition, key) {
            var nextSubStateValue = subStateTransition.stateValue, nextSubActions = subStateTransition.actions, nextSubActivities = subStateTransition.activities;
            if (nextSubActions) {
                if (nextSubActions.onEntry) {
                    (_a = finalActions.onEntry).push.apply(_a, nextSubActions.onEntry);
                }
                if (nextSubActions.actions) {
                    (_b = finalActions.actions).push.apply(_b, nextSubActions.actions);
                }
                if (nextSubActions.onExit) {
                    (_c = finalActions.onExit).push.apply(_c, nextSubActions.onExit);
                }
            }
            if (nextSubActivities) {
                Object.assign(finalActivities, nextSubActivities);
            }
            if (!nextSubStateValue) {
                return stateValue[key];
            }
            return nextSubStateValue;
            var _a, _b, _c;
        });
        return {
            stateValue: finalStateValue,
            actions: finalActions,
            activities: finalActivities
        };
    };
    StateNode.prototype.next = function (event, fullState, history, extendedState) {
        var _this = this;
        var eventType = utils_1.getEventType(event);
        var actionMap = { onEntry: [], onExit: [], actions: [] };
        var activityMap = {};
        if (this.onExit) {
            actionMap.onExit = this.onExit;
        }
        if (this.activities) {
            this.activities.forEach(function (activity) {
                activityMap[utils_1.getEventType(activity)] = false;
                actionMap.onExit = actionMap.onExit.concat(actions_1.stop(activity));
            });
        }
        if (!this.on || !this.on[eventType]) {
            return {
                stateValue: undefined,
                actions: actionMap,
                activities: activityMap
            };
        }
        var transition = this.on[eventType];
        var nextStateString;
        if (typeof transition === 'string') {
            nextStateString = transition;
        }
        else {
            var candidates = Array.isArray(transition)
                ? transition
                : Object.keys(transition).map(function (key) { return (__assign({}, transition[key], { target: key })); });
            for (var _i = 0, candidates_1 = candidates; _i < candidates_1.length; _i++) {
                var candidate = candidates_1[_i];
                var _a = candidate, cond = _a.cond, stateIn = _a.in, transitionActions = _a.actions;
                var extendedStateObject = extendedState || {};
                var eventObject = actions_1.toEventObject(event);
                var isInState = stateIn
                    ? matchesState_1.matchesState(utils_1.toStateValue(stateIn), utils_1.path(this.path.slice(1, -2))(fullState.value))
                    : true;
                if ((!cond || cond(extendedStateObject, eventObject)) &&
                    (!stateIn || isInState)) {
                    nextStateString = candidate.target;
                    if (transitionActions) {
                        actionMap.actions = actionMap.actions.concat(transitionActions);
                    }
                    break;
                }
            }
        }
        if (!nextStateString) {
            return {
                stateValue: undefined,
                actions: actionMap,
                activities: activityMap
            };
        }
        var nextStatePath = utils_1.toStatePath(nextStateString);
        var currentState = this.parent;
        var currentHistory = history;
        var currentPath = this.key;
        nextStatePath.forEach(function (subPath) {
            if (!currentState || !currentState.states) {
                throw new Error("Unable to read '" + subPath + "'");
            }
            if (subPath === HISTORY_KEY) {
                if (Object.keys(currentState.states).length == 0) {
                    subPath = NULL_EVENT;
                }
                else if (currentHistory) {
                    subPath =
                        typeof currentHistory === 'object'
                            ? Object.keys(currentHistory)[0]
                            : currentHistory;
                }
                else if (currentState.initial) {
                    subPath = currentState.initial;
                }
                else {
                    throw new Error("Cannot read '" + HISTORY_KEY + "' from state '" + currentState.id + "': missing 'initial'");
                }
            }
            else if (subPath === NULL_EVENT) {
                actionMap.onExit = [];
                currentState = currentState.states[_this.key];
                return;
            }
            if (subPath !== NULL_EVENT) {
                currentState = currentState.states[subPath];
            }
            if (currentState === undefined) {
                throw new Error("Event '" + event + "' on state '" + currentPath + "' leads to undefined state '" + nextStatePath.join(STATE_DELIMITER) + "'.");
            }
            if (currentState.onEntry) {
                actionMap.onEntry = actionMap.onEntry.concat(currentState.onEntry);
            }
            if (currentState.activities) {
                currentState.activities.forEach(function (activity) {
                    activityMap[utils_1.getEventType(activity)] = true;
                    actionMap.onEntry = actionMap.onEntry.concat(actions_1.start(activity));
                });
            }
            currentPath = subPath;
            if (currentHistory) {
                currentHistory = currentHistory[subPath];
            }
        });
        if (!currentState) {
            throw new Error('no state');
        }
        while (currentState.initial) {
            if (!currentState || !currentState.states) {
                throw new Error("Invalid initial state");
            }
            currentState = currentState.states[currentState.initial];
            if (currentState.onEntry) {
                actionMap.onEntry = actionMap.onEntry.concat(currentState.onEntry);
            }
            if (currentState.activities) {
                currentState.activities.forEach(function (activity) {
                    activityMap[utils_1.getEventType(activity)] = true;
                    actionMap.onEntry = actionMap.onEntry.concat(actions_1.start(activity));
                });
            }
        }
        return {
            stateValue: currentState.getRelativeValue(this.parent),
            actions: actionMap,
            activities: activityMap
        };
    };
    Object.defineProperty(StateNode.prototype, "resolvedStateValue", {
        get: function () {
            var key = this.key;
            if (this.parallel) {
                return _a = {},
                    _a[key] = utils_1.mapValues(this.states, function (stateNode) { return stateNode.resolvedStateValue[stateNode.key]; }),
                    _a;
            }
            if (!this.initial) {
                // If leaf node, value is just the state node's key
                return key;
            }
            return _b = {},
                _b[key] = this.states[this.initial].resolvedStateValue,
                _b;
            var _a, _b;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(StateNode.prototype, "initialStateValue", {
        get: function () {
            var initialStateValue = this.__cache.initialState ||
                (this.parallel
                    ? utils_1.mapValues(this.states, function (state) { return state.initialStateValue; })
                    : typeof this.resolvedStateValue === 'string'
                        ? undefined
                        : this.resolvedStateValue[this.key]);
            this.__cache.initialState = initialStateValue;
            return this.__cache.initialState;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(StateNode.prototype, "initialState", {
        get: function () {
            var initialStateValue = this.initialStateValue;
            if (!initialStateValue) {
                throw new Error("Cannot retrieve initial state from simple state '" + this.id + ".'");
            }
            var entryActions = this.getStateNodes(initialStateValue).reduce(function (actions, stateNode) {
                return stateNode.onEntry ? actions.concat(stateNode.onEntry) : actions;
            }, []);
            return new State_1.State(initialStateValue, undefined, entryActions);
        },
        enumerable: true,
        configurable: true
    });
    StateNode.prototype.getStates = function (stateValue) {
        var _this = this;
        if (typeof stateValue === 'string') {
            return [this.states[stateValue]];
        }
        var stateNodes = [];
        Object.keys(stateValue).forEach(function (key) {
            stateNodes.push.apply(stateNodes, _this.states[key].getStates(stateValue[key]));
        });
        return stateNodes;
    };
    StateNode.prototype.getState = function (relativeStateId) {
        var statePath = utils_1.toStatePath(relativeStateId);
        try {
            return statePath.reduce(function (subState, subPath) {
                if (!subState.states) {
                    throw new Error("Cannot retrieve subPath '" + subPath + "' from node with no states");
                }
                return subState.states[subPath];
            }, this);
        }
        catch (e) {
            throw new Error("State '" + relativeStateId + " does not exist on machine '" + this.id + "'");
        }
    };
    Object.defineProperty(StateNode.prototype, "events", {
        get: function () {
            if (this.__cache.events) {
                return this.__cache.events;
            }
            var states = this.states;
            var events = new Set(this.on ? Object.keys(this.on) : undefined);
            if (states) {
                Object.keys(states).forEach(function (stateId) {
                    var state = states[stateId];
                    if (state.states) {
                        for (var _i = 0, _a = state.events; _i < _a.length; _i++) {
                            var event_1 = _a[_i];
                            events.add("" + event_1);
                        }
                    }
                });
            }
            return (this.__cache.events = Array.from(events));
        },
        enumerable: true,
        configurable: true
    });
    StateNode.prototype.getRelativeValue = function (toNode) {
        var memoizedRelativeValue = toNode
            ? this.__cache.relativeValue.get(toNode)
            : undefined;
        if (memoizedRelativeValue) {
            return memoizedRelativeValue;
        }
        var initialStateValue = this.initialStateValue;
        var relativeValue = initialStateValue
            ? (_a = {},
                _a[this.key] = initialStateValue,
                _a) : this.key;
        var currentNode = this.parent;
        while (currentNode && currentNode !== toNode) {
            var currentInitialState = currentNode.initialStateValue;
            relativeValue = (_b = {},
                _b[currentNode.key] = currentNode.parallel &&
                    typeof currentInitialState === 'object' &&
                    typeof relativeValue === 'object'
                    ? __assign({}, currentInitialState, relativeValue) : relativeValue,
                _b);
            currentNode = currentNode.parent;
        }
        this.__cache.relativeValue.set(toNode, relativeValue);
        return relativeValue;
        var _a, _b;
    };
    return StateNode;
}());
exports.StateNode = StateNode;
function Machine(config) {
    return new StateNode(config);
}
exports.Machine = Machine;
